//=======================================================================
//                        Project Polyphony
//
// File:
//   fm_dispatch_io.v
//
// Abstract:
//   slave access dispatcher
//
// Created:
//   16 March 2009
//
//======================================================================
//
// Copyright (c) 2013, Kenji Ishimaru
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//  -Redistributions of source code must retain the above copyright notice,
//   this list of conditions and the following disclaimer.
//  -Redistributions in binary form must reproduce the above copyright notice,
//   this list of conditions and the following disclaimer in the documentation
//   and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
// EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
//  Revision History
//  29 June 2013: microblaze I/O support
//
//  microblaze memory mapping
//
//  address(byte)                      address(/32bit)
//
//  0xffff_ffff +------------------+ 0x00ff_ffff
//              |  (reserved)      |
//  0xf000_0000 +------------------+ 0x00c0_0000
//              |  LPDDR           |
//  0xe000_0000 +------------------+ 0x0080_0000
//              |  System Register |         
//  0xd000_0000 +------------------+ 0x0040_0000
//              |  USB             |
//  0xc000_0000 +------------------+ 0x0000_0000
//  (256MB each)
//
//
// Revision History
// $Date: 2014-04-03 13:36:47 +0900 (Thu, 03 Apr 2014) $
// $Rev: 51 $
//
// synthesis attribute keep_hierarchy of fm_dispatcher_io is yes;
// synthesis attribute register_balancing of fm_dispatcher_io is no;

module fm_dispatch_io (
    clk_core,
    rst_x,
    // local interface
    i_req,
    i_wr,
    i_adrs,
    o_ack,
    i_be,
    i_wd,
    o_rstr,
    o_rd,
    // usb wishbone
    o_usb_strobe,
    o_usb_we,
    o_usb_adrs,
    o_usb_wd,
    i_usb_rd,
    i_usb_ack,
    // system register access
    o_sys_strobe,
    o_sys_wr,
    o_sys_adrs,
    o_sys_wd,
    o_sys_be,
    i_sys_ack,
    i_sys_rstr,
    i_sys_rd,
    // ddr access
    o_ddr_strobe,
    o_ddr_wr,
    o_ddr_adrs,
    o_ddr_wd,
    o_ddr_be,
    i_ddr_ack,
    i_ddr_rstr,
    i_ddr_rd
);
//////////////////////////////////
// I/O port definition
//////////////////////////////////
    input           clk_core;
    input           rst_x;
    // local interface
    input           i_req;
    input           i_wr;
    input  [31:0]   i_adrs;
    output          o_ack;
    input  [3:0]    i_be;
    input  [31:0]   i_wd;
    output          o_rstr;
    output [31:0]   o_rd;
    // usb wishbone
    output          o_usb_strobe;
    output          o_usb_we;
    output [7:0]    o_usb_adrs;
    output [7:0]    o_usb_wd;
    input  [7:0]    i_usb_rd;
    input           i_usb_ack;
    // system register access
    output o_sys_strobe;
    output o_sys_wr;
    output [21:0] o_sys_adrs;
    output [31:0] o_sys_wd;
    output [3:0] o_sys_be;
    input i_sys_ack;
    input i_sys_rstr;
    input [31:0] i_sys_rd;
    // ddr access
    output o_ddr_strobe;
    output o_ddr_wr;
    output [21:0] o_ddr_adrs;
    output [31:0] o_ddr_wd;
    output [3:0] o_ddr_be;
    input i_ddr_ack;
    input i_ddr_rstr;
    input [31:0] i_ddr_rd;

//////////////////////////////////
// reg
//////////////////////////////////
    reg             r_rstr;
    reg    [31:0]   r_rd;
//////////////////////////////////
// wire
//////////////////////////////////
    wire            w_hit_usb;
    wire            w_hit_sys;
    wire            w_hit_ddr;
    wire            w_ack;
    wire            w_rstr;
    wire    [31:0]  w_rd;
//////////////////////////////////
// assign
//////////////////////////////////
    // address decode
    assign w_hit_usb = (i_adrs[31:28] == 4'hc);
    assign o_usb_strobe = i_req & w_hit_usb;
    assign o_usb_we = i_wr;
    assign o_usb_adrs = i_adrs[9:2];
    assign o_usb_wd = i_wd[7:0];

    assign w_hit_sys = (i_adrs[31:28] == 4'hd);
    assign o_sys_strobe = i_req & w_hit_sys;
    assign o_sys_wr = i_wr;
    assign o_sys_adrs = i_adrs[23:2];
    assign o_sys_wd = i_wd;
    assign o_sys_be = i_be;

    assign w_hit_ddr = (i_adrs[31:28] == 4'he);
    assign o_ddr_strobe = i_req & w_hit_ddr;
    assign o_ddr_wr = i_wr;
    assign o_ddr_adrs = i_adrs[23:2];
    assign o_ddr_wd = i_wd;
    assign o_ddr_be = i_be;

    assign w_ack = (o_sys_strobe) ? i_sys_ack :
                   (o_ddr_strobe) ? i_ddr_ack :
                                    i_usb_ack;
    assign w_rd = (i_sys_rstr) ? i_sys_rd :
                  (i_ddr_rstr) ? i_ddr_rd :
                                 i_usb_rd;
    assign w_rstr = i_sys_rstr | i_ddr_rstr | 
                    (o_usb_strobe & !o_usb_we & i_usb_ack);
    assign o_ack = w_ack;
    assign o_rstr = r_rstr;
    assign o_rd = r_rd;

//////////////////////////////////
// always
//////////////////////////////////
    always @(posedge clk_core or negedge rst_x) begin
        if (~rst_x) begin
            r_rstr <= 1'b0;
        end else begin
            r_rstr <= w_rstr;
        end
    end

    always @(posedge clk_core) begin
        r_rd <= w_rd;
    end

endmodule
